import numpy as np
def rmse(label, target):
    err = target - label
    value = np.sqrt(np.sum(np.square(err))/label.size)
    return value

def psnr(label, target):
    minv = np.min(label)
    maxv = np.max(label)
    sca = 255.0 / (maxv - minv)
    ln = (label - minv) * sca
    tn = (target - minv) * sca
    rmv = rmse(ln, tn)
    value = 10 * np.log((255.0**2)/(rmv**2))/np.log(10)
    return value


'''
The function to compute SSIM
@param param: img_mat_1 1st 2D matrix
@param param: img_mat_2 2nd 2D matrix
'''
def ssim(label, target):
    minv = np.min(label)
    maxv = np.max(label)
    sca = 255.0 / (maxv - minv)
    img_mat_1 = (label - minv) * sca
    img_mat_2 = (target - minv) * sca
    import numpy
    import scipy.ndimage
    from numpy.ma.core import exp
    from scipy.constants.constants import pi
    #Variables for Gaussian kernel definition
    gaussian_kernel_sigma=1.5
    gaussian_kernel_width=11
    gaussian_kernel=numpy.zeros((gaussian_kernel_width,gaussian_kernel_width))
    
    #Fill Gaussian kernel
    for i in range(gaussian_kernel_width):
        for j in range(gaussian_kernel_width):
            gaussian_kernel[i,j]=\
            (1/(2*pi*(gaussian_kernel_sigma**2)))*\
            exp(-(((i-5)**2)+((j-5)**2))/(2*(gaussian_kernel_sigma**2)))

    #Convert image matrices to double precision (like in the Matlab version)
    img_mat_1=img_mat_1.astype(numpy.float)
    img_mat_2=img_mat_2.astype(numpy.float)
    
    #Squares of input matrices
    img_mat_1_sq=img_mat_1**2
    img_mat_2_sq=img_mat_2**2
    img_mat_12=img_mat_1*img_mat_2
    
    #Means obtained by Gaussian filtering of inputs
    img_mat_mu_1=scipy.ndimage.filters.convolve(img_mat_1,gaussian_kernel)
    img_mat_mu_2=scipy.ndimage.filters.convolve(img_mat_2,gaussian_kernel)
        
    #Squares of means
    img_mat_mu_1_sq=img_mat_mu_1**2
    img_mat_mu_2_sq=img_mat_mu_2**2
    img_mat_mu_12=img_mat_mu_1*img_mat_mu_2
    
    #Variances obtained by Gaussian filtering of inputs' squares
    img_mat_sigma_1_sq=scipy.ndimage.filters.convolve(img_mat_1_sq,gaussian_kernel)
    img_mat_sigma_2_sq=scipy.ndimage.filters.convolve(img_mat_2_sq,gaussian_kernel)
    
    #Covariance
    img_mat_sigma_12=scipy.ndimage.filters.convolve(img_mat_12,gaussian_kernel)
    
    #Centered squares of variances
    img_mat_sigma_1_sq=img_mat_sigma_1_sq-img_mat_mu_1_sq
    img_mat_sigma_2_sq=img_mat_sigma_2_sq-img_mat_mu_2_sq
    img_mat_sigma_12=img_mat_sigma_12-img_mat_mu_12;
    
    #c1/c2 constants
    #First use: manual fitting
    c_1=6.5025
    c_2=58.5225
    
    #Second use: change k1,k2 & c1,c2 depend on L (width of color map)
    l=255
    k_1=0.01
    c_1=(k_1*l)**2
    k_2=0.03
    c_2=(k_2*l)**2
    
    #Numerator of SSIM
    num_ssim=(2*img_mat_mu_12+c_1)*(2*img_mat_sigma_12+c_2)
    #Denominator of SSIM
    den_ssim=(img_mat_mu_1_sq+img_mat_mu_2_sq+c_1)*\
    (img_mat_sigma_1_sq+img_mat_sigma_2_sq+c_2)
    #SSIM
    ssim_map=num_ssim/den_ssim
    index=numpy.average(ssim_map)

    return index



"""
Metrics of many samples
"""


def mean_square_error(labels, targets):
    from .transform import unbatch
    from .metrics import mean_square_error as msem
    labels = unbatch(labels)
    targets = unbatch(targets)
    results = [msem(msem) for l, t in zip(labels, targets)]
    return results


def analysis(labels, targets=None, *, stats=('mean_square_error',)):
    pass


def get_stats(name):
    if name.lower() in ['mean_square_error', 'mse']:
        return mean_square_error

def stat(tensor):
    import numpy as np
    return {'min': np.min(tensor), 'max': np.max(tensor), 'mean': np.mean(tensor), 'std': np.std(tensor), 'shape': list(tensor.shape)}

def compare(label, target):
    from .metrics import rmse, psnr, ssim
    return rmse(label, target), psnr(label, target), ssim(label, target)


import click
import warnings


def get_main_net(net_name):
    warnings.warn(DeprecationWarning("Deprecated, no suggestion."))
    from dxpy.learn.net.zoo.srms.train import main as srms
    if net_name == "srms":
        return srms


def get_main_func(net_name):
    warnings.warn(DeprecationWarning("Deprecated, no suggestion."))
    pass


@click.command()
@click.option('--config', '-c', type=str, help='configs .yml filename')
@click.option('--net', '-n', type=str, help='name of main net', default=None)
def train(net, config):
    warnings.warn(DeprecationWarning("Deprecated, use train2."))
    main_net_train = get_main_net(net)
    main_net_train()


@click.command()
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--name', '-n', type=str, help='cluster name of dataset in config file, e.g. cluster/dataset/task0.')
def dataset(name, config):
    from .dataset import dataset_dist
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config(config)
    dataset_dist(name)


@click.command()
@click.option('--name', '-n', type=str, help='name of network')
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--job_name', '-j', type=str)
@click.option('--task_index', '-t', type=int)
@click.option('--run', '-r', type=str, help='run task', default='train')
def main(name, job_name, task_index, run, config):
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config(config)
    if name == 'sin':
        from .zoo.sin.main import main
        main(run, job_name, task_index)
    elif name == 'srms':
        from .zoo.srms.main import main
        main(run, job_name, task_index)
    else:
        raise ValueError("Unknown name {} for dxpy.ln.main CLI.".format(name))


@click.command()
@click.option('--train_config_name', '-t', type=str)
@click.option('--dataset_maker_name', '-d', type=str)
@click.option('--network_maker_name', '-n', type=str)
@click.option('--summary_maker_name', '-s', type=str)
@click.option('--config', '-c', type=str, default='dxln.yml')
def train2(train_config_name, config, dataset_maker_name=None, network_maker_name=None, summary_maker_name=None):
    if dataset_maker_name is None:
        dataset_maker_name = train_config_name
    if network_maker_name is None:
        network_maker_name = train_config_name
    if summary_maker_name is None:
        summary_maker_name = train_config_name
    from .base import DxlnRunEnvrionment
    from .train import train_task_local
    with DxlnRunEnvrionment(config):
        train_task_local(dataset_maker_name,
                         network_maker_name,
                         summary_maker_name)


@click.command()
@click.option('--cluster_file', '-f', type=str, default='cluster.yml', help='cluster config file name.')
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--job_name', '-j', type=str)
@click.option('--task_index', '-t', type=int)
def train_dist(cluster_file, job_name, task_index, config):
    """
    Run dist tasks
    """
    from dxpy.learn.distribute.cluster import get_cluster_spec
    from .base import DxlnRunEnvrionment
    cluster = get_cluster_spec(cluster_file)
    from .train import train_task_dist
    with DxlnRunEnvrionment(config):
        train_task_dist(name='cluster/{}/task{}'.format(job_name, task_index),
                        cluster=cluster)

@click.command()
@click.option('--task', '-t', type=str)
@click.option('--dataset', '-d', type=str)
@click.option('--output', '-o', type=str, default='result.npz')
@click.option('--nb_samples', '-n', type=int, default=0)
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--recon_method', '-r', type=str, default='fbp')
def infer(task, dataset, output, nb_samples, config, recon_method):
    from .inference import infer_sino_sr, recon_sino, infer_phan_sr, infer_mct
    from .base import DxlnRunEnvrionment
    with DxlnRunEnvrionment(config, with_pre_work=True):
        if task == 'sinosr':
            infer_sino_sr(dataset, nb_samples, output)
        if task == 'phansr':
            infer_phan_sr(dataset, nb_samples, output)
        if task == 'reconsr':
            recon_sino(dataset, nb_samples, output, recon_method)
        if task == 'mctsr':
            infer_mct(dataset, nb_samples, output)
            





import click

@click.command()
def hello():
    click.echo('HELLO FROM SR.HELLO!')

@click.command()
@click.option('--input', '-i', type=str)
@click.option('--output', '-o', type=str)
def analysis(input, output):
    from .analysis import analysis_result
    analysis_result(input, output)



COLUMNS = ['IDX', 'TYPE', 'NB_DOWN', 'RMSE', 'PSNR', 'SSIM']

def analysis_one_image(label, target, name):    
    from dxpy.tensor.metrics import psnr, rmse, ssim
    image_type = name[:-2]
    nb_down = int(name[-2:-1])
    return (image_type, nb_down, rmse(label, target), psnr(label, target), ssim(label, target))

def analysis_result(filename, output):
    import pandas as pd
    from tqdm import tqdm
    import numpy as np
    results = np.load(filename)
    nb_images = results['phantom'].shape[0]
    anas = []
    for idx in tqdm(range(nb_images)):
        for k in results.keys():
            if k == 'phantom':
                continue
            anas.append([idx] + list(analysis_one_image(results['phantom'][idx, ...], results[k][idx, ...], name=k)))
    df = pd.DataFrame(data=anas, columns=COLUMNS)
    df.to_csv(output)


from ..graph import Graph
from ..model.image import MultiDownSampler
from dxpy.configs import configurable
from ..config import config


class SuperResolutionDataset(Graph):
    """
    SuperResolutionDataset is a special kind of dataset which
    """
    @configurable(config, with_name=True)
    def __init__(self, name, origial_dataset_maker, input_key=None, nb_down_sample=None, with_shape_info=None, origin_shape=None, **config):
        super().__init__(name, input_key=input_key,
                         nb_down_sample=nb_down_sample,
                         with_shape_info=with_shape_info,
                         origin_shape=origin_shape,
                         **config)
        self._dataset_maker = origial_dataset_maker
        self.__construct()

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({'with_shape_info': False}, super()._default_config())

    def __down_sample_keys(self, start):
        return [('image{}x'.format(2**i), 2**i) for i in range(start, self.param('nb_down_sample') + 1)]

    def __construct(self):
        self.register_node('dataset',
                           self._dataset_maker())
        down_sample_ratios = {k: [v] * 2 for k,
                              v in self.__down_sample_keys(1)}
        origin_key = self.__down_sample_keys(0)[0][0]
        self.register_node('down_sampler',
                           MultiDownSampler(self.nodes['dataset'][self.param('input_key')],
                                            down_sample_ratios=down_sample_ratios,
                                            keep_original=True, original_key=origin_key,
                                            register_output_with_prefix=False))
        for k, _ in self.__down_sample_keys(0):
            self.register_node(k, self.nodes['down_sampler'][k])


def random_crop_multiresolution(images, target_shape_high_resolution, nb_down_sample=None):
    """
    Args:
        images: dict-like object. Containing 'image{0,1,2,3,...}x'.
    """
    from dxpy.learn.utils.tensor import shape_as_list
    import tensorflow as tf
    if nb_down_sample is None:
        nb_down_sample = len(images) - 1

    with tf.name_scope('random_crop_multi_resolution'):
        shapes = [shape_as_list(images['image{}x'.format(i)])
                  for i in range(nb_down_sample + 1)]
        down_sample_ratios = [sh // sl for sh, sl in zip(shapes[0], shapes[1])]
        for i in range(len(shapes) - 1):
            for sh, sl, dr in zip(shapes[i], shapes[i + 1], down_sample_ratios):
                if sl * dr != sh:
                    raise ValueError(
                        "Invalid inputs shapes: {}".format(shapes))
        for sh, st in zip(shapes[0], target_shape_high_resolution):
            if sh < st:
                raise ValueError("Can not perfrom random crop on inputs with shape {} with target shape {}.".format(
                    shapes, target_shape_high_resolution))
        target_shapes = []
        for i in range(nb_down_sample + 1):
            target_shapes.append([s // (dr**i)
                                  for s, dr in zip(target_shape_high_resolution, down_sample_ratios)])
        max_offsets = [sl - tl for sl,
                       tl in zip(shapes[-1], target_shapes[-1])]
        offset = []
        for mo in max_offsets:
            if mo == 0:
                offset.append(0)
            else:
                offset.append(tf.random_uniform([], 0, mo, dtype=tf.int64))
        offsets = [offset]
        for i in range(nb_down_sample):
            offset_new = [o * dr for o,
                          dr in zip(offsets[-1], down_sample_ratios)]
            offsets.append(offset_new)
        offsets = list(offsets[::-1])
        results = dict()
        for i in range(nb_down_sample + 1):
            k = 'image{}x'.format(i)
            results[k] = tf.slice(images[k], offsets[i], target_shapes[i])
        return results



import os
import tensorflow as tf
from dxpy.configs import configurable, ConfigsView
import numpy as np
from dxpy.learn.graph import Graph, NodeKeys
from dxpy.learn.config import config

DEFAULT_FILE_NAME = 'mCT.npy'
NB_IMAGES = 983


def data_type_np(key):
    if key == 'id':
        return np.int64
    else:
        return np.float32


def data_type_tf(key):
    if key == 'id':
        return tf.int64
    else:
        return tf.float32


def data_shape(key):
    return {
        'phantom': [192, 192],
        'sinogram1x': [768 * 2, 384],
        'sinogram2x': [384 * 2, 192],
        'sinogram4x': [192 * 2, 96],
        'sinogram8x': [96 * 2, 48],
        'id': [],
    }[key]


@configurable(ConfigsView(config, 'datasets/mCT_sinograms'))
def _npy_file(path: str, npy_file_name=DEFAULT_FILE_NAME):
    from dxpy.core.path import Path
    return str(Path(path) / npy_file_name)


def _load_sample(idx, data):
    result = dict()
    result['phantom'] = data[idx][4]
    for i in range(4):
        result['sinogram{}x'.format(2**i)] = data[idx][i].T
    result['id'] = idx
    return result


def _processing(result):
    from dxpy.tensor.transform import rotate
    for k in result:
        if k == 'id':
            continue
        if not k == 'phantom':
            sino = result[k]
            sino_d = np.flip(sino, axis=0)
            sino_d = rotate(sino_d, -1, 0)
            result[k] = np.concatenate([sino, sino_d], axis=1)
            result[k] = result[k].T
            result[k] = np.concatenate([result[k]] * 2, axis=0)
        result[k] = result[k].astype(data_type_np(k))
        # if k == 'phantom':
            # result[k] = result[k] / np.sum(result[k]) * 1e6
        # if k == 'phantom':
            # result[k] = result[k] / np.sum(result[k]) * 4e6
    return result


def dataset_generator(ids):
    data = np.load(_npy_file())
    for idx in ids:
        result = _load_sample(idx, data)
        result = _processing(result)
        yield result


class Dataset(Graph):
    MEAN = 9.93 
    STD = 7.95
    BASE_SHAPE = [384, 384]

    @configurable(config, with_name=True)
    def __init__(self, name='mCT_sinograms', batch_size=32, ids=None, shuffle=True, dataset_type='train', **kw):
        super().__init__(name=name, batch_size=batch_size,
                         shuffle=shuffle, dataset_type=dataset_type, **kw)
        if ids is None:
            if dataset_type == 'train':
                self._ids = list(range(0, int(NB_IMAGES * 0.8)))
            elif dataset_type == 'test':
                self._ids = list(range(int(NB_IMAGES * 0.8), NB_IMAGES))
            elif dataset_type == 'full':
                self._ids = list(range(NB_IMAGES))
        else:
            self._ids = ids
        self._keys = ['id', 'phantom'] + \
            ['sinogram{}x'.format(2**i) for i in range(4)]

        dataset = self._create_dataset()
        dataset = self._processing_dataset(dataset)
        self._register_dataset(dataset)

    def _create_dataset(self):
        from functools import partial
        output_types = {k: data_type_tf(k) for k in self._keys}
        output_shapes = {k: data_shape(k) for k in self._keys}
        dataset_generator_partial = partial(dataset_generator, ids=self._ids)
        return tf.data.Dataset.from_generator(dataset_generator_partial,
                                              output_types,
                                              output_shapes)

    def _format_tensors(self, data):
        return {k: tf.reshape(tf.cast(data[k], data_type_tf(k)), data_shape(k)) for k in data}

    def _processing_dataset(self, dataset):
        dataset = (dataset
                   .repeat()
                   .map(self._format_tensors))
        if self.param('shuffle'):
            dataset = dataset.shuffle(1024)
        return dataset.batch(self.param('batch_size'))

    def _register_dataset(self, dataset):
        from ...utils.tensor import ensure_shape
        iterator = dataset.make_one_shot_iterator()
        next_element = iterator.get_next()
        # self.register_main_node(next_element)
        for k in self._keys:
            shape = [self.param('batch_size')] + list(data_shape(k)) + [1]
            self.register_node(k, ensure_shape(next_element[k], shape=shape))



import os
import tensorflow as tf
from dxpy.configs import configurable, ConfigsView
import numpy as np
from dxpy.learn.graph import Graph, NodeKeys
from dxpy.learn.config import config

DEFAULT_FILE_NAME = 'pet_rebin2.h5'
NB_IMAGES = 4770
CROP_OFFSET = 32


def data_type_np(key):
    if key == 'id':
        return np.int64
    else:
        return np.float32


def data_type_tf(key):
    if key == 'id':
        return tf.int64
    else:
        return tf.float32


def data_shape(key):
    return {
        'sinogram': [640, 128 - 2 * CROP_OFFSET],
        'id': [],
    }[key]


@configurable(ConfigsView(config, 'datasets/mice_sinograms'))
def _h5_file(path: str, file_name=DEFAULT_FILE_NAME):
    from dxpy.core.path import Path
    return str(Path(path) / file_name)


def _load_sample(idx, data):
    result = {'id': idx, 'sinogram': data[idx, ...].T}
    return result


def _processing(result):
    import numpy as np
    result['sinogram'] = np.concatenate([result['sinogram']] * 2)[:, CROP_OFFSET:-CROP_OFFSET]
    return result


def dataset_generator(ids):
    import h5py
    with h5py.File(str(_h5_file())) as fin:
        data = np.array(fin['sinograms'])
        data = np.maximum(data, 0.0)
    for idx in ids:
        result = _load_sample(idx, data)
        result = _processing(result)
        yield result


class Dataset(Graph):
    MEAN = 0.217
    STD = 0.993
    HIGH_NOISE_MAX_DEPTH = 128
    DEPTH_PER_MICE = 159

    @configurable(config, with_name=True)
    def __init__(self, name='mice_sinograms', batch_size=32, ids=None, shuffle=True, dataset_type='train', **kw):
        super().__init__(name=name,
                         batch_size=batch_size,
                         shuffle=shuffle,
                         dataset_type=dataset_type,
                         **kw)
        if ids is None:
            if dataset_type == 'train':
                self._ids = list(range(0, int(NB_IMAGES * 0.8)))
            elif dataset_type == 'test':
                self._ids = list(range(int(NB_IMAGES * 0.8), NB_IMAGES))
            elif dataset_type == 'full':
                self._ids = list(range(NB_IMAGES))
        else:
            self._ids = ids
        self._ids = list(filter(lambda x: x % self.DEPTH_PER_MICE <
                                self.HIGH_NOISE_MAX_DEPTH, self._ids))
        self._keys = ['id', 'sinogram']

        dataset = self._create_dataset()
        dataset = self._processing_dataset(dataset)
        self._register_dataset(dataset)

    def _create_dataset(self):
        from functools import partial
        output_types = {k: data_type_tf(k) for k in self._keys}
        output_shapes = {k: data_shape(k) for k in self._keys}
        dataset_gen = partial(dataset_generator, ids=self._ids)
        return tf.data.Dataset.from_generator(dataset_gen, output_types, output_shapes)

    def _format_tensors(self, data):
        return {k: tf.reshape(tf.cast(data[k], data_type_tf(k)), data_shape(k)) for k in data}

    def _processing_dataset(self, dataset):
        dataset = (dataset
                   .repeat()
                   .cache()
                   .map(self._format_tensors))
        if self.param('shuffle'):
            dataset = dataset.shuffle(1024)
        return dataset.batch(self.param('batch_size'))

    def _register_dataset(self, dataset):
        from ...utils.tensor import ensure_shape
        iterator = dataset.make_one_shot_iterator()
        next_element = iterator.get_next()
        # self.register_main_node(next_element)
        for k in self._keys:
            shape = [self.param('batch_size')] + list(data_shape(k)) + [1]
            self.register_node(k, ensure_shape(next_element[k], shape=shape))



import click

class CLI(click.MultiCommand):
    """
    Helper of projects specified CLI module.
    """
    commands = {'sr': None}

    def __init__(self):
        super().__init__(name='pjs', help=__class__.__doc__)

    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        from .super_resolution.cli import main as main_sr
        mapped = {'sr': main_sr}
        if name in self.commands and self.commands[name] is None:
            self.commands[name] = mapped[name]
        return self.commands.get(name)

main = CLI()



import numpy as np
from typing import List, TypeVar, Tuple


def random_crop_offset(input_shape, target_shape, *, batched=False):
    max_offset = [s - t for s, t in zip(input_shape, target_shape)]
    if any(map(lambda x: x < 0, max_offset)):
        raise ValueError("Invalid input_shape {} or target_shape {}.".format(
            input_shape, target_shape))
    if not batched:
        offset = []
        for s in max_offset:
            if s == 0:
                offset.append(0)
            else:
                offset.append(np.random.randint(0, s))
        return offset
    else:
        offsets = []
        if max_offset[0] > 0:
            raise ValueError(
                "Random crop offset input_shape[0] and target_shape[0]")


def unstack_auto(tensor, *, hint: str='image'):
    pass


def unstack(tensor, axis: int=0, size: int=1, *,  squeeze: bool=True):
    """
    Unstack tensor along given axies with given sizes. 
    Args:
    - tensor: np.ndarray
    - axies: int
    - sizes: int
    - squeeze: bool, whether squeeze dim when size equals one.

    Returns:
    - List[tensor]
    """
    indices = range(0, tensor.shape[axis], size)
    splitted = np.split(tensor, indices)
    if squeeze and size == 1:
        sli = [slice(0, s) for s in tensor.shape]
        sli[axis] = 0
        splitted = [t[sli] for t in splitted]
    return splitted


def mirror(tensor, axis, *, forward: bool=True):
    tensor_m = np.flip(tensor, axis)
    if forward:
        return np.concatenate([tensor, tensor_m], axis=axis)
    else:
        return np.concatenate([tensor_m, tensor], axis=axis)



# def pad(tensor, axis: int, target_size: int, offset: int=0, method: str='constant', *, padding_value=0)
#     def padding_one_block(source, target, idx, ids, method, mirror_axis=None):
#         if method is None:
#             target[idx] = source[ids]
#         elif method.lower() == 'constant':
#             target[idx] = constant_value
#         elif method.lower() == 'period':
#             target[idx] = source[ids]
#         elif method.lower() == 'mirror':
#             ids[mirror_axis] = slice(ids[mirror_axis].stop - 1, None, -1)
#             target[idx] = source[ids]

#     def index(tensor, target_size, axis, offset):
#         ids = list()
#         idt = list()
#         for i in range(tensor.ndim):
#             if i == axis:
#                 max_idx = target_size
#                 slice_min = min([max([offset, 0]), max_idx])

#                 slice_max = min(
#                     [max([offset + tensor.shape[i], 0]), max_idx])
#             else:
#                 slice_min = 0
#                 slice_max = tensor.shape[i]
#             idt.append(slice(slice_min, slice_max))
#             ids.append(slice(0, slice_max - slice_min))
#         return idt, ids

#     def current_method(method, is_mirror):
#         if method == 'mirror':
#             if is_mirror:
#                 c_method = 'mirror'
#                 is_mirror = False
#             else:
#                 c_method = 'period'
#                 is_mirror = True
#         else:
#             c_method = method
#         return c_method, is_mirror

#     offset_start = offset
#     offset_now = offset_start
#     next_shape = list(tensor.shape)
#     next_shape[axis] = target_size
#     result = np.zeros(next_shape)
#     is_mirror = False
#     while offset_now < target_size:
#         idt, ids = index(tensor, target_size, i, offset_now)
#         c_method, is_mirror = current_method(method, is_mirror)
#         padding_one_block(tensor, result, idt, ids, c_method, axis)
#         offset_now += tensor.shape[axis]
#     is_mirror = True
#     offset_now = offset_start - tensor.shape[axis]
#     while offset_now > - tensor.shape[axis]:
#         idt, ids = index(tensor, target_size, i, offset_now)
#         c_method, is_mirror = current_method(method, is_mirror)
#         padding_one_block(tensor, result, idt, ids, c_method, axis)
#         offset_now -= tensor.shape[axis]
#     return result

# def padding()
    # TODO: add guard for tf.Tensor
    # tensor = np.array(tensor)
    # tensors = [tensor]
    # for i in range(len(axies)):
    #     tensors = [_unbatch_single_axis(t) for t in tensors]
    #     for j, v in enumerate(axies):
    #         if v > axies:
    #             axies[j] = v - 1
    # return tensors
    # if isinstance(tensor, np.ndarray):
    # return list(map(lambda x: x[0, ...], np.split(tensor, tensor.shape[0])))
#    raise TypeError("numpy.ndarray is required, got {}.".format(type(tensor)))


def batch(tensors):
    return np.array(tensors)


def maybe_unbatch(tensors: TypeVar('T', np.ndarray, List[np.ndarray])) -> List[np.ndarray]:
    if isinstance(tensors, (list, tuple)):
        return tensors
    else:
        return unbatch(tensors)


def padding(tensor: np.ndarray, target_shape: List[int], offset: TypeVar('T', int, List[int])=0, method: TypeVar('T', str, List[str])=None, *, padding_order: List[int]=None, constant_value: float=None):
    """
    Supported method:
        period: periodically padding tensor, assuming period equals to tensor shape.
        mirror: mirror reflect
        constant: pad with constant value
    """
    for i, v in enumerate(target_shape):
        if v is None:
            target_shape[i] = tensor.shape[i]
    if isinstance(offset, int):
        offset = [offset] * tensor.ndim
    if isinstance(method, str):
        method = [method] * tensor.ndim
    if padding_order is None:
        padding_order = list(range(tensor.ndim))

    def padding_one_block(source, target, idx, ids, method, mirror_axis=None):
        if method is None:
            target[idx] = source[ids]
        elif method.lower() == 'constant':
            target[idx] = constant_value
        elif method.lower() == 'period':
            target[idx] = source[ids]
        elif method.lower() == 'mirror':
            ids[mirror_axis] = slice(ids[mirror_axis].stop - 1, None, -1)
            target[idx] = source[ids]

    def index(tensor, target_size, axis, offset):
        ids = list()
        idt = list()
        for i in range(tensor.ndim):
            if i == axis:
                max_idx = target_size
                slice_min = min([max([offset, 0]), max_idx])

                slice_max = min(
                    [max([offset + tensor.shape[i], 0]), max_idx])
            else:
                slice_min = 0
                slice_max = tensor.shape[i]
            idt.append(slice(slice_min, slice_max))
            ids.append(slice(0, slice_max - slice_min))
        return idt, ids

    def current_method(method, is_mirror):
        if method == 'mirror':
            if is_mirror:
                c_method = 'mirror'
                is_mirror = False
            else:
                c_method = 'period'
                is_mirror = True
        else:
            c_method = method
        return c_method, is_mirror

    def padding_one_dim(tensor, axis, target_size, offset, method):
        offset_start = offset
        offset_now = offset_start
        next_shape = list(tensor.shape)
        next_shape[axis] = target_size
        result = np.zeros(next_shape)
        is_mirror = False
        while offset_now < target_size:
            idt, ids = index(tensor, target_size, i, offset_now)
            c_method, is_mirror = current_method(method, is_mirror)
            padding_one_block(tensor, result, idt, ids, c_method, axis)
            offset_now += tensor.shape[axis]
        is_mirror = True
        offset_now = offset_start - tensor.shape[axis]
        while offset_now > - tensor.shape[axis]:
            idt, ids = index(tensor, target_size, i, offset_now)
            c_method, is_mirror = current_method(method, is_mirror)
            padding_one_block(tensor, result, idt, ids, c_method, axis)
            offset_now -= tensor.shape[axis]
        return result
    result = np.array(tensor)
    for i in padding_order:
        if method[i] is None:
            continue
        result = padding_one_dim(
            result, i, target_shape[i], offset[i], method[i])
    return result


def rotate(tensor, offset, axis):
    slis = [slice(0, s) for s in tensor.shape]
    slil = list(slis)
    slir = list(slis)
    offset = offset % tensor.shape[axis]
    slil[axis] = slice(offset, tensor.shape[axis])
    slir[axis] = slice(0, offset)
    tl = tensor[slil]
    tr = tensor[slir]
    result = np.concatenate([tl, tr], axis=axis)
    return result



class DxlnRunEnvrionment:
    def __init__(self, config_file_name='dxln.yml', with_pre_work=False):
        self._config = config_file_name
        self._with_pre_work = with_pre_work
        pass

    def __enter__(self):
        from dxpy.learn.utils.general import pre_work, load_yaml_config
        load_yaml_config(self._config)
        if self._with_pre_work:
            pre_work()
        return self

    def __exit__(self, type, value, trackback):
        pass



import tensorflow as tf
from dxpy.configs import configurable, ConfigsView
# from ..config import get_configs_view, config
from dxpy.learn.config import config
from .base import NodeKeys
import warnings


def mean_square_error(label, data):
    with tf.name_scope('mean_squared_error'):
        return tf.sqrt(tf.reduce_mean(tf.square(label - data)))


def l1_error(label, data):
    with tf.name_scope('l1_error'):
        return tf.reduce_mean(tf.abs(label - data))

@configurable(ConfigsView(config).get('poission_loss'))
def poission_loss(label, data, *, compute_full_loss=False):
    with tf.name_scope('poission_loss'):
        label = tf.maximum(label, 0.0)
        data = tf.maximum(data, 0.0)
        # return log_possion_loss(tf.log(label), data)
        return tf.reduce_mean(tf.keras.losses.poisson(label, data))


def log_possion_loss(log_label, data, *, compute_full_loss=False):
    """
    log_label: log value of expectation (inference)
    data: Poission sample
    """
    with tf.name_scope('log_poission_loss'):
        data = tf.maximum(data, 0.0)
        return tf.reduce_mean(tf.nn.log_poisson_loss(data, log_label, compute_full_loss))


from ..model.base import Model


class PoissionLossWithDenorm(Model):
    """
    Inputs:
    NodeKeys.INPUT: infer
    NodeKeys.LABEL: label
    Outputs:
    NodeKeys.OUTPUT: scalar loss
    """
    @configurable(config, with_name=True)
    def __init__(self, name, inputs, with_log=False, threshold=10, mean=0.0, std=1.0, **kw):
        warnings.warn(DeprecationWarning(
            "PoissionLossWithDenorm is deprecated, replace it with poission_loss or get_loss_func('poi'), and place denorm into main network."))
        super().__init__(name, inputs=inputs, with_log=with_log,
                         threshold=threshold, mean=mean, std=std, **kw)

    def _kernel(self, feeds):
        from dxpy.debug.utils import dbgmsg
        dbgmsg(self.param('mean'))
        dbgmsg(self.param('std'))

        label = feeds[NodeKeys.LABEL]
        infer = feeds[NodeKeys.INPUT]
        with tf.name_scope('denorm_white'):
            label = label * \
                tf.constant(self.param('std')) + \
                tf.constant(self.param('mean'))
            infer = infer * \
                tf.constant(self.param('std')) + \
                tf.constant(self.param('mean'))
        if self.param('with_log'):
            with tf.name_scope('denorm_log_for_data'):
                infer = tf.exp(infer)
            with tf.name_scope('loss'):
                loss = log_possion_loss(label, infer)
        else:
            with tf.name_scope('loss'):
                loss = poission_loss(label, infer)
        return {NodeKeys.MAIN: loss}


def get_loss_func(name):
    if name.lower() in ['mse', 'mean_square_error', 'l2']:
        return mean_square_error
    if name == 'poi':
        return poission_loss
    if name == 'l1':
        return l1_error
    raise ValueError("Unknown error name {}.".format(name))


class CombinedSupervisedLoss(Model):
    @configurable(config, with_name=True)
    def __init__(self, name, inputs, loss_names, loss_weights, **kw):
        super().__init__(name, inputs=inputs, loss_names=loss_names,
                         loss_weights=loss_weights, **kw)

    def _kernel(self, feeds):
        if not NodeKeys.LABEL in feeds:
            raise ValueError("{} is required for feeds of {}, but not found in feeds".format(
                NodeKeys.LABEL, __class__))
        if not NodeKeys.INPUT in feeds:
            raise ValueError("{} is required for feeds of {}, but not found in feeds".format(
                NodeKeys.INPUT, __class__))
        label = feeds[NodeKeys.LABEL]
        data = feeds[NodeKeys.INPUT]
        with tf.name_scope('combined_losses'):
            losses = []
            for n, w in zip(self.param('loss_names'), self.param('loss_weights')):
                losses.append(get_loss_func(n)(label, data) * tf.constant(w))
            with tf.name_scope('summation'):
                loss = tf.add_n(losses)
        result = {NodeKeys.OUTPUT: loss}
        for i, n in enumerate(self.param('loss_names')):
            result.update({'loss/' + n: losses[i]})
        return result



from dxpy.configs import configurable
from dxpy.learn.config import config
from dxpy.learn.train.summary_2 import SummaryWriter
from dxpy.learn.graph import NodeKeys
import tensorflow as tf


class SinSummaryWriter(SummaryWriter):
    @configurable(config, with_name=True)
    def __init__(self, network, name='summary', **kw):
        summary_tensors = {
            'loss': network[NodeKeys.LOSS],
        }
        super().__init__(name=name, tensors=summary_tensors, nb_runs=dict(), **kw)




def create_global_scalars():
    from . import global_global_step
    from . import global_keep_prob
    import tensorflow as tf
    # global_global_step.create()
    tf.train.get_or_create_global_step()
    global_keep_prob.create()



import click
from tqdm import tqdm


def infer_sino_sr(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    dbgmsg(dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]
    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(
            tf.float32, tensor_shape(k)) for k in keys}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_STD = 9.27
    STAT_MEAN = 9.76
    BASE_SHAPE = (640, 320)

    dataset_configs = config_view['dataset']['srms']
    with_noise = dataset_configs['with_poission_noise']
    if with_noise:
        PREFIX = 'input'
    else:
        PREFIX = 'label'

    def crop_sinogram(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        o1 = (tensor.shape[0] - target_shape[0]) // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:-o1, o2:-o2]

    def run_infer(idx):
        input_key = '{}/image{}x'.format(PREFIX, 2**nb_down_sample)
        low_sino = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        low_sino = (low_sino - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_sino}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_sinogram(inf)
        itpc = crop_sinogram(itp)
        infc = infc * STAT_STD + STAT_MEAN
        itpc = itpc * STAT_STD + STAT_MEAN
        return infc, itpc

    phans = []
    sino_highs = []
    sino_lows = []
    sino_itps = []
    sino_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        sino_highs.append(crop_sinogram(
            data_raw['{}/image1x'.format(PREFIX)][idx, :, :]))
        sino_lows.append(crop_sinogram(data_raw['{}/image{}x'.format(
            PREFIX, 2**nb_down_sample)][idx, ...], [s // (2**nb_down_sample) for s in BASE_SHAPE]))
        sino_inf, sino_itp = run_infer(idx)
        sino_infs.append(sino_inf)
        sino_itps.append(sino_itp)

    results = {'phantom': phans, 'sino_itps': sino_itps,
               'sino_infs': sino_infs, 'sino_highs': sino_highs, 'sino_lows': sino_lows}
    np.savez(output, **results)


def infer_phan_sr(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    print('Using dataset file:', dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    dataset_configs = config_view['dataset']['srms']
    with_noise = dataset_configs['with_poission_noise']
    dbgmsg(with_noise)
    if with_noise:
        PREFIX = 'noise'
    else:
        PREFIX = 'clean'

    def data_key(nd):
        return '{}{}x'.format(PREFIX, 2**nd)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]
    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(tf.float32, tensor_shape(data_key(i % 4)))
                   for i, k in enumerate(keys)}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_MEAN = 15.26 
    STAT_STD = 22.0 
    STAT_MEAN_LOW = 15.26 * (4.0**nb_down_sample)
    STAT_STD_LOW = 22.0 * (4.0**nb_down_sample)
    BASE_SHAPE = (256, 256)

    def crop_image(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        o1 = (tensor.shape[0] - target_shape[0]) // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:-o1, o2:-o2]

    input_key = data_key(nb_down_sample)
    dbgmsg('input_key:', input_key)

    def run_infer(idx):
        low_phan = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        # low_phan = (low_phan - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_phan}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_image(inf)
        itpc = crop_image(itp)
        infc = infc * STAT_STD_LOW + STAT_MEAN_LOW
        itpc = itpc * STAT_STD_LOW + STAT_MEAN_LOW
        return infc, itpc

    phans = []
    img_highs = []
    img_lows = []
    img_itps = []
    img_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        img_high = crop_image(data_raw[data_key(0)][idx, :, :])
        img_high = img_high * STAT_STD + STAT_MEAN
        # img_high = img_high * STAT_STD / \
        # (4.0**nb_down_sample) + STAT_MEAN / (4.0**nb_down_sample)
        img_highs.append(img_high)
        img_low = crop_image(data_raw[data_key(nb_down_sample)][idx, ...],
                             [s // (2**nb_down_sample) for s in BASE_SHAPE])
        img_low = img_low * STAT_STD_LOW + STAT_MEAN_LOW
        img_lows.append(img_low)
        img_inf, img_itp = run_infer(idx)
        img_infs.append(img_inf)
        img_itps.append(img_itp)
    
    img_highs = np.array(img_highs)
    img_infs = np.array(img_infs) / (4.0**nb_down_sample)
    img_itps = np.array(img_itps) / (4.0**nb_down_sample)
    img_lows = np.array(img_lows) / (4.0**nb_down_sample)

    results = {'phantom': phans, 'img_itps': img_itps,
               'img_infs': img_infs, 'img_highs': img_highs, 'img_lows': img_lows}
    np.savez(output, **results)


def infer_mct(dataset, nb_samples, output):
    """
    Use network in current directory as input for inference
    """
    import tensorflow as tf
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network
    from dxpy.configs import ConfigsView
    from dxpy.learn.config import config
    import numpy as np
    import yaml
    from dxpy.debug.utils import dbgmsg
    print('Using dataset file:', dataset)
    data_raw = np.load(dataset)
    data_raw = {k: np.array(data_raw[k]) for k in data_raw.keys()}
    config_view = ConfigsView(config)

    def data_key(nd):
        return 'image{}x'.format(2**nd)

    def tensor_shape(key):
        shape_origin = data_raw[key].shape
        return [1] + list(shape_origin[1:3]) + [1]

    with tf.name_scope('inputs'):
        keys = ['input/image{}x'.format(2**i) for i in range(4)]
        keys += ['label/image{}x'.format(2**i) for i in range(4)]
        dataset = {k: tf.placeholder(tf.float32, tensor_shape(data_key(i % 4)))
                   for i, k in enumerate(keys)}

    network = get_network('network/srms', dataset=dataset)
    nb_down_sample = network.param('nb_down_sample')
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.train.MonitoredTrainingSession(
        checkpoint_dir='./save', config=config, save_checkpoint_secs=None)

    STAT_MEAN = 9.93 
    STAT_STD = 7.95
    STAT_MEAN_LOW = 9.93 * (4.0**nb_down_sample)
    STAT_STD_LOW = 7.95 * (4.0**nb_down_sample)
    BASE_SHAPE = (384, 384)

    def crop_image(tensor, target_shape=None):
        if target_shape is None:
            target_shape = BASE_SHAPE
        if len(tensor.shape) == 4:
            tensor = tensor[0, :, :, 0]
        # o1 = (tensor.shape[0] - target_shape[0]) // 2
        o1 = tensor.shape[0] // 2
        o2 = (tensor.shape[1] - target_shape[1]) // 2
        return tensor[o1:o1+target_shape[0], o2:-o2]

    input_key = data_key(nb_down_sample)
    dbgmsg('input_key:', input_key)

    def run_infer(idx):
        low_phan = np.reshape(
            data_raw[input_key][idx, :, :], tensor_shape(input_key))
        # low_phan = (low_phan - STAT_MEAN) / STAT_STD
        feeds = {dataset['input/image{}x'.format(2**nb_down_sample)]: low_phan}
        inf, itp = sess.run([network['outputs/inference'],
                             network['outputs/interp']], feed_dict=feeds)
        infc = crop_image(inf)
        itpc = crop_image(itp)
        infc = infc * STAT_STD_LOW + STAT_MEAN_LOW
        itpc = itpc * STAT_STD_LOW + STAT_MEAN_LOW
        return infc, itpc

    phans = []
    img_highs = []
    img_lows = []
    img_itps = []
    img_infs = []
    NB_MAX = data_raw['phantom'].shape[0]
    for idx in tqdm(range(nb_samples), ascii=True):
        if idx > NB_MAX:
            import sys
            print('Index {} out of range {}, stop running and store current result...'.format(
                idx, NB_MAX), file=sys.stderr)
            break

        phans.append(data_raw['phantom'][idx, ...])
        img_high = crop_image(data_raw[data_key(0)][idx, :, :])
        img_high = img_high * STAT_STD + STAT_MEAN
        # img_high = img_high * STAT_STD / \
        # (4.0**nb_down_sample) + STAT_MEAN / (4.0**nb_down_sample)
        img_highs.append(img_high)
        img_low = crop_image(data_raw[data_key(nb_down_sample)][idx, ...],
                             [s // (2**nb_down_sample) for s in BASE_SHAPE])
        img_low = img_low * STAT_STD_LOW + STAT_MEAN_LOW
        img_lows.append(img_low)
        img_inf, img_itp = run_infer(idx)
        img_infs.append(img_inf)
        img_itps.append(img_itp)
    
    img_highs = np.array(img_highs)
    img_infs = np.array(img_infs) / (4.0**nb_down_sample)
    img_itps = np.array(img_itps) / (4.0**nb_down_sample)
    img_lows = np.array(img_lows) / (4.0**nb_down_sample)

    results = {'phantom': phans, 'sino_itps': img_itps,
               'sino_infs': img_infs, 'sino_highs': img_highs, 'sino_lows': img_lows}
    np.savez(output, **results)

def recon_sino(sinograms_filename, nb_samples, output, recon_method):
    import numpy as np
    from dxpy.medical_image_processing.phantom import Phantom2DSpec
    from dxpy.medical_image_processing.reconstruction.parallel import reconstruction2d
    from dxpy.medical_image_processing.detector import Detector2DParallelRing
    sinograms = np.load(sinograms_filename)
    sino_highs = np.array(sinograms['sino_highs'])
    sino_lows = np.array(sinograms['sino_lows'])
    sino_infs = np.array(sinograms['sino_infs'])
    sino_itps = np.array(sinograms['sino_itps'])
    phans = np.array(sinograms['phantom'])
    nb_views_high = sino_highs.shape[1]
    nb_views_low = sino_lows.shape[1]
    nb_views_high //= 2
    nb_views_low //= 2
    sino_highs = sino_highs[:, :nb_views_high, :]
    sino_lows = sino_lows[:, :nb_views_low, :]
    sino_infs = sino_infs[:, :nb_views_high, :]
    sino_itps = sino_itps[:, :nb_views_high, :]
    sino_lows = sino_lows * nb_views_high**2 / (nb_views_low**2)
    phan_spec = Phantom2DSpec(shape=[256] * 2)
    detec_high = Detector2DParallelRing(views=np.linspace(
        0, np.pi, nb_views_high, endpoint=False), nb_sensors=nb_views_high, sensor_width=1.0)
    detec_low = Detector2DParallelRing(views=np.linspace(
        0, np.pi, nb_views_low, endpoint=False), nb_sensors=nb_views_low, sensor_width=1.0 * nb_views_high / nb_views_low)
    recon_highs = []
    recon_lows = []
    recon_infs = []
    recon_itps = []

    def recon_kernel(sinogram, detec):
        sinogram = np.maximum(sinogram, 0.0)
        if recon_method == 'fbp':
            recon = reconstruction2d(sinogram, detec, phan_spec)
        elif recon_method == 'sart':
            recon = reconstruction2d(
                sinogram, detec, phan_spec, method='SART_CUDA', iterations=100)
        recon = np.maximum(recon, 0.0)
        recon = recon / np.sum(recon) * 1e6
        return recon

    for i in tqdm(range(nb_samples), ascii=True):
        recon_highs.append(recon_kernel(sino_highs[i, ...], detec_high))
        recon_lows.append(recon_kernel(sino_lows[i, ...], detec_low))
        recon_infs.append(recon_kernel(sino_infs[i, ...], detec_high))
        recon_itps.append(recon_kernel(sino_itps[i, ...], detec_high))
    results = {'phantom': phans, 'recon_highs': recon_highs,
               'recon_lows': recon_lows, 'recon_infs': recon_infs, 'recon_itps': recon_itps}
    np.savez(output, **results)



from dxpy.learn.dataset.api import get_dataset
from dxpy.learn.net.api import get_network
from dxpy.learn.session import SessionDist
from tqdm import tqdm
from dxpy.learn.graph import NodeKeys
import time
import arrow
from dxpy.learn.scalar import global_step
import tensorflow as tf

def main_worker(task_index, cluster):
    from dxpy.learn.distribute.cluster import get_server
    from dxpy.learn.utils.general import pre_work
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    server = get_server(cluster, 'worker', task_index, config=config)
    with tf.device(tf.train.replica_device_setter(worker_device="/job:worker/task:{}".format(task_index), cluster=cluster)):
        pre_work()
        dataset = get_dataset(name='dataset/srms')
        network = get_network(name='network/srms', dataset=dataset)
        result = network()
        hooks = [tf.train.StepCounterHook()]
        with tf.train.MonitoredTrainingSession(master=server.target,
                                                is_chief=(task_index == 0),
                                                checkpoint_dir="./save",
                                                config=config,
                                                hooks=hooks) as sess:
            from dxpy.learn.session import set_default_session
            set_default_session(sess)
            network.nodes['trainer'].run('set_learning_rate')
            for _ in tqdm(range(10000000000000), ascii=True):
                network.train()

def main_ps(task_index, cluster):
    server = get_server(cluster, 'ps', task_index)
    server.join()

def main_dataset(task_index, cluster, join=False):
    from dxpy.learn.distribute.dataset import get_dist_dataset
    server = get_server(cluster, 'dataset', task_index)
    dataset = get_dist_dataset(name='cluster/dataset/task{}'.format(task_index)) 
    if join:
        server.join()
    return dataset

def main_summary(task_index, cluster, dataset):
    from dxpy.learn.distribute.cluster import get_server
    from dxpy.learn.net.zoo.srms.summ import SRSummaryWriter_v2
    from dxpy.learn.utils.general import pre_work
    from dxpy.learn.session import set_default_session
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    server = get_server(cluster, 'worker', task_index, config=config)
    with tf.device(tf.train.replica_device_setter(worker_device="/job:worker/task:{}".format(task_index), cluster=cluster)):
        pre_work()
        dataset = get_dataset(name='dataset/srms')
        network = get_network(name='network/srms', dataset=dataset)
        result = network()
        hooks = [tf.train.StepCounterHook()]
        with tf.train.MonitoredTrainingSession(master=server.target,
                                               is_chief=(task_index == 0),
                                               checkpoint_dir="./save",
                                               config=config,
                                               hooks=hooks) as sess:
            set_default_session(sess)
            network.nodes['trainer'].run('set_learning_rate')
            sw = SRSummaryWriter_v2(network=network, tensors=tensors, name=summary_config)
            sw.post_session_created()
            while True:
                sw.auto_summary()

def main(task='train', job_name='worker', task_index=0, cluster_config='cluster.yml'):
    from dxpy.learn.distribute.cluster import get_cluster_spec, get_server, get_nb_tasks
    from dxpy.learn.distribute.dataset import get_dist_dataset
    from dxpy.learn.distribute.ps import get_dist_network as get_ps_network
    from dxpy.learn.distribute.worker import apply_dist_network
    from dxpy.learn.distribute.summary import get_dist_summary
    from dxpy.learn.distribute.worker import get_dist_network as get_worker_network
    from dxpy.learn.utils.general import pre_work
    cluster = get_cluster_spec(cluster_config, job_name=None)
    if not job_name == 'worker':
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        server = get_server(cluster, job_name, task_index, config=config)
        datasets = []
        nb_datasets = get_nb_tasks(cluster_config, 'dataset')
        for i in range(1):
            datasets.append(get_dist_dataset(
                name='cluster/dataset/task{}'.format(i)))
    if job_name == 'dataset':
        server.join()
        return
    elif job_name == 'ps':
        server.join()
        return
    elif job_name == 'worker':
        main_worker(task_index, cluster)
    elif job_name in ['summary']:
        with tf.device(tf.train.replica_device_setter(worker_device="/job:worker/task:{}".format(task_index), cluster=cluster)):
            pre_work()
            network = get_network(name='network/srms',
                                  dataset=datasets[task_index % nb_datasets])
            result = network()
        if job_name == 'worker':
            hooks = [tf.train.StepCounterHook()]
            with tf.train.MonitoredTrainingSession(master=server.target,
                                                   is_chief=(task_index == 0),
                                                   #    is_chief=True,
                                                   checkpoint_dir="./save",
                                                   config=config,
                                                   hooks=hooks) as sess:
                from dxpy.learn.session import set_default_session
                set_default_session(sess)
                network.nodes['trainer'].run('set_learning_rate')
            # sess = SessionDist(target=server.target)
                # with sess._sess.as_default():
                for _ in tqdm(range(10000000000000), ascii=True):
                    network.train()
    if job_name == 'summary':
        name = 'cluster/summary/task{}'.format(task_index)
        result = apply_dist_network(
            network=network, dataset=datasets[-1], name=name)
        result.update(datasets[-1].nodes)
        sw = get_dist_summary(tensors=result, network=network, name=name)
        hooks = [tf.train.StepCounterHook()]
        with tf.train.MonitoredTrainingSession(master=server.target,
                                               is_chief=False,
                                               config=config,
                                               hooks=hooks) as sess:
            from dxpy.learn.session import set_default_session
            set_default_session(sess)
            sw.post_session_created()
            while True:
                sw.auto_summary()
            #
        # sess = SessionDist(target=server.target)
        # with sess.as_default():
        #     sw.post_session_created()
        #     while True:
        #         sw.auto_summary()
        return



"""
Type converters for tensors
"""
import numpy as np


def rescale_to_uint16(arr):
    """
    Convert tensor to UINT16 Type while keeping maximum information.
    """
    UINT16MAX = 65535
    if isinstance(arr, np.ndarray):
        arr = np.maximum(arr, 0.0)
        arr = arr / np.max(arr) * UINT16MAX
        arr = np.minimum(arr, UINT16MAX)
        return arr.astype(np.uint32)
    else:
        import tensorflow as tf
        if isinstance(arr, tf.Tensor):
            raise NotImplementedError
        else:
            raise TypeError(
                "Requires np.yndarray or tf.Tensor, got {}.".format(type(arr)))



"""
Super resolution dataset for analytical phantom sinogram dataset
"""
from typing import List

import tensorflow as tf

from dxpy.configs import configurable

from ...config import config
from ...graph import Graph, NodeKeys
from dxpy.learn.utils.tensor import shape_as_list
import numpy as np


class AnalyticalPhantomSinogramDatasetForSuperResolution(Graph):
    @configurable(config, with_name=True)
    def __init__(self, name='datasets/apssr',
                 image_type: str='sinogram',
                 nb_down_sample: int=3,
                 *,
                 log_scale: bool=False,
                 with_white_normalization: bool=True,
                 with_poission_noise: bool=False,
                 with_noise_label: bool=False,
                 with_phase_shift: bool=False,
                 target_shape: List[int]=None,
                 low_dose: bool= False,
                 low_dose_ratio: float= 10.0,
                 **kw):
        """
        Args:
            -   image_type: 'sinogram' or 'image'
            -   log_scale: produce log datas
            -   with_white_normalization: normalize by mean and std
            -   with_poission_noise (apply to sinogram only): perform poission sample
            -   target_shape: (apply to sinogram only), random crop shape
        Returns:
            a `Graph` object, which has several nodes:
        Raises:
        """
        super().__init__(name, image_type=image_type,
                         log_scale=log_scale,
                         with_poission_noise=with_poission_noise,
                         target_shape=target_shape,
                         with_white_normalization=with_white_normalization,
                         with_noise_label=with_noise_label,
                         with_phase_shift=with_phase_shift,
                         low_dose=low_dose,
                         low_dose_ratio=low_dose_ratio,
                         nb_down_sample=nb_down_sample, **kw)
        # from ...model.normalizer.normalizer import FixWhite, ReduceSum
        # from ...model.tensor import ShapeEnsurer
        # from dxpy.core.path import Path
        # name = Path(name)
        # if image_type == 'sinogram':
        #     fields = ['sinogram', 'id']
        # elif image_type == 'sinogram_with_phantom':
        #     fields = ['sinogram', 'phantom', 'id']
        # else:
        #     fields = ['phantom', 'id'] + ['recon{}x'.format(2**i)
        #                                   for i in range(self.param('nb_down_sample') + 1)]
        # with tf.name_scope('aps_{img_type}_dataset'.format(img_type=image_type)):
        #     dataset_origin = Dataset(
        #         self.name / 'analytical_phantom_sinogram', fields=fields)
        #     tid = dataset_origin['id']
        #     if image_type in ['sinogram', 'sinogram_with_phantom']:
        #         dataset = self._process_sinogram(dataset_origin)
        #     else:
        #         dataset = self._process_recons(dataset_origin)
        # # for k in dataset
        #     # self.register_node(k, dataset[k]):

        # # Old fashion
        # for i in range(self.param('nb_down_sample') + 1):
        #     img_key = 'image{}x'.format(2**i)
        #     self.register_node('noise/' + img_key, dataset['noise/' + img_key])
        #     self.register_node('clean/' + img_key, dataset['clean/' + img_key])
        #     self.register_node('input/image{}x'.format(2**i),
        #                        dataset['input/image{}x'.format(2**i)])
        #     if self.param('image_type') == 'image' and i == 0:
        #         self.register_node('label/image1x', dataset['label/phantom'])
        #         continue
        #     if self.param('with_noise_label'):
        #         self.register_node('label/image{}x'.format(2**i),
        #                            dataset['input/image{}x'.format(2**i)])
        #     else:
        #         self.register_node('label/image{}x'.format(2**i),
        #                            dataset['label/image{}x'.format(2**i)])

        if image_type in ['sinogram', 'sinogram_with_phantom', 'mice_sinograms']:
            if image_type in ['sinogram', 'sinogram_with_phantom']:
                result = self._get_sinogram_aps()
            else:
                result = self._get_sinogram_mice()
            if self.param('with_poission_noise'):
                input_keys = [
                    'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            else:
                input_keys = [
                    'clean/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            if self.param('with_noise_label'):
                label_keys = [
                    'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            else:
                label_keys = [
                    'clean/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        elif image_type == 'recons_full_size':
            result = self._get_phantom_aps()
            input_keys = [
                'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            label_keys = [
                'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        elif image_type == 'mCT':
            result = self._get_sinogram_mct()
            input_keys = [
                'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            label_keys = [
                'noise/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        elif image_type == 'recon_ms':
            result = self._get_recons_ms()
            if self.param('with_poission_noise'):
                input_keys = [
                    'noise{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            else:
                input_keys = [
                    'clean{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            if self.param('with_noise_label'):
                label_keys = [
                    'noise{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
            else:
                label_keys = [
                    'clean{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        else:
            raise ValueError('Unknown image type {}.'.format(image_type))
        output_input_keys = [
            'input/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        output_label_keys = [
            'label/image{}x'.format(2**i) for i in range(self.param('nb_down_sample') + 1)]
        for ki, ko in zip(input_keys, output_input_keys):
            self.register_node(ko, result[ki])
        for ki, ko in zip(label_keys, output_label_keys):
            self.register_node(ko, result[ki])

    def _get_sinogram_aps(self):
        from ..raw.analytical_phantom_sinogram import Dataset
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import SuperResolutionDataset
        fields = ['sinogram', 'id', 'phantom']
        with tf.name_scope('aps_sinogram_dataset'):
            dataset = Dataset(
                self.name / 'analytical_phantom_sinogram', fields=fields)
            self.register_node('id', dataset['id'])
            self.register_node('phantom', dataset['phantom'])
            if self.param('log_scale'):
                stat = dataset.LOG_SINO_STAT
            else:
                stat = dataset.SINO_STAT
            dataset = dataset['sinogram']
            if self.param('with_poission_noise'):
                with tf.name_scope('add_with_poission_noise'):
                    noise = tf.random_poisson(dataset, shape=[])
                    dataset = tf.concat([noise, dataset], axis=0)
                if self.param('log_scale'):
                    dataset = tf.log(dataset + 0.4)
            if self.param('with_white_normalization'):
                dataset = FixWhite(name=self.name / 'fix_white',
                                   inputs=dataset, mean=stat['mean'], std=stat['std']).as_tensor()
            dataset = tf.random_crop(dataset,
                                     [shape_as_list(dataset)[0]] + list(self.param('target_shape')) + [1])
            dataset = SuperResolutionDataset(self.name / 'super_resolution',
                                             lambda: {'image': dataset},
                                             input_key='image',
                                             nb_down_sample=self.param('nb_down_sample'))
            keys = ['image{}x'.format(2**i)
                    for i in range(dataset.param('nb_down_sample') + 1)]
            result = dict()
            if self.param('with_poission_noise'):
                result.update(
                    {'noise/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys})
                result.update(
                    {'clean/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
            else:
                result.update({'clean/' + k: dataset[k] for k in keys})
                result.update({'noise/' + k: dataset[k] for k in keys})
        return result

    def _get_recons_ms(self):
        from ..raw.analytical_phantom_sinogram import Dataset
        from dxpy.learn.model.normalizer import FixWhite
        ratios = [2**i for i in range(self.param('nb_down_sample') + 1)]
        fields = ['phantom', 'id'] + ['clean{}x'.format(r) for r in ratios] + [
            'noise{}x'.format(r) for r in ratios]
        with tf.name_scope('dataset_recon_ms'):
            dataset = Dataset(
                self.name / 'analytical_phantom_sinogram', fields=fields)
            self.register_node('id', dataset['id'])
            self.register_node('phantom', dataset['phantom'])
            keys_noise = ['noise{}x'.format(r) for r in ratios]
            keys_clean = ['clean{}x'.format(r) for r in ratios]
            stat = dataset.RECON_MS_STAT
            cleans = dict()
            noises = dict()
            if self.param('with_white_normalization'):
                with tf.name_scope('normalizatoin_clean'):
                    for i, k in enumerate(keys_clean):
                        cleans[k] = FixWhite(name=self.name / 'fix_white', inputs=dataset[k],
                                             mean=stat['mean'] * (4.0)**i, std=stat['std'] * (4.0)**i).as_tensor()
                with tf.name_scope('normalizatoin_noise'):
                    for i, k in enumerate(keys_noise):
                        noises[k] = FixWhite(name=self.name / 'fix_white', inputs=dataset[k],
                                             mean=stat['mean'] * (4.0)**i, std=stat['std'] * (4.0)**i).as_tensor()
            # for i, r in ratios:
            #     self.register_node('clean/image{}x'.format(r), cleans[keys_clean[i]])
            #     self.register_node('noise/image{}x'.format(r), noises[keys_noise[i]])
            # if self.param('with_poission_noise'):
            #     images = {'input/image{}x'.format(r): noises['noise{}x'.format(r)] for r in ratios}
            # else:
            #     images = {'input/image{}x'.format(r): cleans['clean{}x'.format(r)] for r in ratios}
            # if self.param('with_noise_lable'):
            #     labels = {'label/image{}x'.format(r): noises['noise{}x'.format(r)] for r in ratios}
            # else:
            #     labels = {'label/image{}x'.format(r): cleans['clean{}x'.format(r)] for r in ratios}
            result = dict()
            result.update(cleans)
            result.update(noises)
            return result

    def _get_phantom_aps(self):
        from ..raw.analytical_phantom_sinogram import Dataset
        from ..super_resolution import SuperResolutionDataset
        fields = ['phantom', 'id'] + ['recon{}x'.format(2**i)
                                      for i in range(self.param('nb_down_sample') + 1)]
        with tf.name_scope('aps_phantom_dataset'):
            dataset_origin = Dataset(
                self.name / 'analytical_phantom_sinogram', fields=fields)
            self.register_node('id', dataset_origin['id'])
            self.register_node('phantom', dataset['phantom'])
            label = dataset['phantom']
            keys = ['recon{}x'.format(2**i)
                    for i in range(self.param('nb_down_sample') + 1)]
            dataset = dataset_origin
            if self.param('log_scale'):
                stat = dataset.LOG_RECON_STAT
            else:
                stat = dataset.RECON_STAT
            dataset = {k: dataset[k] for k in keys}
            for i, k in enumerate(keys):
                if i == 0:
                    continue
                dataset[k] = tf.nn.avg_pool(dataset[k],
                                            [1] + [2**i, 2**i] + [1],
                                            padding='SAME',
                                            strides=[1] + [2**i, 2**i] + [1]) * (2.0**i * 2.0**i)
            if self.param('log_scale'):
                for k in keys:
                    dataset[k] = tf.log(dataset[k] + 1.0)
                label = tf.log(label + 1.0)
            if self.param('with_white_normalization'):
                for i, k in enumerate(keys):
                    dataset[k] = FixWhite(name=self.name / 'fix_white' / k,
                                          inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()
                label = FixWhite(name=self.name / 'fix_white' / 'phantom',
                                 inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()
            result = dict()
            for i, k in enumerate(keys):
                result['noise/image{}x'.format(2**i)] = dataset[k]
            result['label/phantom'] = phantom
            return result

    def _get_sinogram_mct(self):
        from ..raw.mCT import Dataset
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import random_crop_multiresolution
        with tf.variable_scope('mCT_dataset'):
            dataset_origin = Dataset(self.name / 'mCT')
            self.register_node('id', dataset_origin['id'])
            self.register_node('phantom', dataset_origin['phantom'])
            dataset = dict()
            keys = ['noise/image{}x'.format(2**i)
                    for i in range(self.param('nb_down_sample') + 1)]
            for i, k in enumerate(keys):
                with tf.variable_scope('normalization_{}x'.format(2**i)):
                    fwn = FixWhite(self.name / 'normalization_{}x'.format(2**i),
                                   inputs=dataset_origin['sinogram{}x'.format(
                                       2**i)],
                                   mean=dataset_origin.MEAN * (4.0**i),
                                   std=dataset_origin.STD * (4.0**i))
                    dataset['noise/image{}x'.format(2**i)] = fwn.as_tensor()
                    # dataset[k] = dataset_origin['sinogram{}x'.format(2**i)]
            crop_keys = ['image{}x'.format(i) for i in range(len(keys))]
            images = {'image{}x'.format(
                i): dataset[k] for i, k in enumerate(keys)}
            images_cropped = random_crop_multiresolution(images,
                                                         [shape_as_list(images['image1x'])[0]] + list(self.param('target_shape')) + [1])

            result = dict()
            for ki, ko in zip(crop_keys, keys):
                result[ko] = images_cropped[ki]
            return result

    def _get_sinogram_mice(self):
        from ..raw.mice import Dataset
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import SuperResolutionDataset
        with tf.name_scope('mice_sinogram_dataset'):
            dataset = Dataset(self.name / 'mice_sinogram')
            self.register_node('id', dataset['id'])
            bs = dataset.param('batch_size')
            stat = {'mean': dataset.MEAN, 'std': dataset.STD}
            dataset = dataset['sinogram']
            if self.param('with_poission_noise'):
                with tf.name_scope('add_with_poission_noise'):
                    if self.param('low_dose'):
                        ratio = self.param('low_dose_ratio')
                        ratio_norm = 4e6 * bs / ratio
                        dataset = dataset / tf.reduce_sum(dataset) * ratio_norm
                        stat['mean'] = stat['mean'] / ratio
                        stat['std'] = stat['std'] / ratio
                    else:
                        dataset = dataset / tf.reduce_sum(dataset) * 4e6 * bs
                    noise = tf.random_poisson(dataset, shape=[])
                    dataset = tf.concat([noise, dataset], axis=0)
            if self.param('with_white_normalization'):
                dataset = FixWhite(name=self.name / 'fix_white',
                                   inputs=dataset, mean=stat['mean'], std=stat['std']).as_tensor()
            dataset = tf.random_crop(dataset,
                                     [shape_as_list(dataset)[0]] + list(self.param('target_shape')) + [1])
            dataset = SuperResolutionDataset(self.name / 'super_resolution',
                                             lambda: {'image': dataset},
                                             input_key='image',
                                             nb_down_sample=self.param('nb_down_sample'))
            keys = ['image{}x'.format(2**i)
                    for i in range(dataset.param('nb_down_sample') + 1)]
            result = dict()
            if self.param('with_poission_noise'):
                result.update(
                    {'noise/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys})
                result.update(
                    {'clean/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
            else:
                result.update({'clean/' + k: dataset[k] for k in keys})
                result.update({'noise/' + k: dataset[k] for k in keys})
        return result

    def _process_sinogram(self, dataset):
        from ...model.normalizer.normalizer import ReduceSum, FixWhite
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import SuperResolutionDataset
        from ...utils.tensor import shape_as_list
        if self.param('log_scale'):
            stat = dataset.LOG_SINO_STAT
        else:
            stat = dataset.SINO_STAT
        # dataset = ReduceSum(self.name / 'reduce_sum', dataset['sinogram'],
            # fixed_summation_value=1e6).as_tensor()
        if 'phantom' in dataset:
            phan = dataset['phantom']
        else:
            phan = None
        dataset = dataset['sinogram']
        if self.param('with_poission_noise'):
            with tf.name_scope('add_with_poission_noise'):
                noise = tf.random_poisson(dataset, shape=[])
                dataset = tf.concat([noise, dataset], axis=0)
        if self.param('log_scale'):
            dataset = tf.log(dataset + 0.4)
        if self.param('with_white_normalization'):
            dataset = FixWhite(name=self.name / 'fix_white',
                               inputs=dataset, mean=stat['mean'], std=stat['std']).as_tensor()
        # random phase shift
        # if self.param('with_phase_shift'):
        #     phase_view = tf.random_uniform(
        #         [], 0, shape_as_list(dataset)[1], dtype=tf.int64)
        #     dataset_l = dataset[:, phase_view:, :, :]
        #     dataset_r = dataset[:, :phase_view, :, :]
        #     dataset = tf.concat([dataset_l, dataset_r], axis=1)
        dataset = tf.random_crop(dataset,
                                 [shape_as_list(dataset)[0]] + list(self.param('target_shape')) + [1])
        dataset = SuperResolutionDataset(self.name / 'super_resolution',
                                         lambda: {'image': dataset},
                                         input_key='image',
                                         nb_down_sample=self.param('nb_down_sample'))
        keys = ['image{}x'.format(2**i)
                for i in range(dataset.param('nb_down_sample') + 1)]
        if self.param('with_poission_noise'):
            result = {
                'input/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys}
            result.update(
                {'label/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
        else:
            result = {'input/' + k: dataset[k] for k in keys}
            result.update({'label/' + k: dataset[k] for k in keys})
        result.update(
            {'noise/' + k: dataset[k][:shape_as_list(dataset[k])[0] // 2, ...] for k in keys})
        result.update(
            {'clean/' + k: dataset[k][shape_as_list(dataset[k])[0] // 2:, ...] for k in keys})
        if phan is not None:
            result.update({'phantom': phan})
        return result

    def _process_recons(self, dataset):
        from ...model.normalizer.normalizer import ReduceSum, FixWhite
        from dxpy.learn.model.normalizer import FixWhite
        from ..super_resolution import SuperResolutionDataset
        from ...utils.tensor import shape_as_list
        keys = ['recon{}x'.format(2**i)
                for i in range(self.param('nb_down_sample') + 1)]
        if self.param('log_scale'):
            stat = dataset.LOG_RECON_STAT
        else:
            stat = dataset.RECON_STAT
        phantom = dataset['phantom']
        dataset = {k: dataset[k] for k in keys}
        for i, k in enumerate(keys):
            if i == 0:
                continue
            dataset[k] = tf.nn.avg_pool(dataset[k],
                                        [1] + [2**i, 2**i] + [1],
                                        padding='SAME',
                                        strides=[1] + [2**i, 2**i] + [1]) * (2.0**i * 2.0**i)

        # dataset = {k: ReduceSum(self.name / 'reduce_sum' / k, dataset[k], fixed_summation_value=1e6).as_tensor() for k in keys}
        if self.param('log_scale'):
            for k in keys:
                dataset[k] = tf.log(dataset[k] + 1.0)
            phantom = tf.log(phantom + 1.0)
        if self.param('with_white_normalization'):
            for i, k in enumerate(keys):
                dataset[k] = FixWhite(name=self.name / 'fix_white' / k,
                                      inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()
            phantom = FixWhite(name=self.name / 'fix_white' / 'phantom',
                               inputs=dataset[k], mean=stat['mean'], std=stat['std']).as_tensor()

        result = dict()
        for i, k in enumerate(keys):
            result['input/image{}x'.format(2**i)] = dataset[k]
            result['label/image{}x'.format(2**i)
                   ] = result['input/image{}x'.format(2**i)]
        result['label/phantom'] = phantom
        return result



from pprint import pprint

import numpy as np
import tensorflow as tf
from dxpy.learn.session import Session
from dxpy.learn.train.summary import SummaryWriter
from dxpy.learn.utils.general import pre_work
from tqdm import tqdm
from dxpy.learn.config import config
import yaml
from dxpy.configs import configurable, ConfigsView


@configurable(ConfigsView(config).get('train'))
def get_train_configs(summary_freq=1000, save_freq=10000, steps=100000000):
    return {'summary_freq': summary_freq,
            'save_freq': save_freq,
            'steps': steps}


def train(definition_func):
    with open('dxln.yml') as fin:
        ycfg = yaml.load(fin)
    config.update(ycfg)
    pre_work()
    train_cfgs = get_train_configs()
    steps = train_cfgs['steps']
    summary_freq = train_cfgs['summary_freq']
    save_freq = train_cfgs['save_freq']
    network, summary = definition_func(ycfg)
    session = Session()
    with session.as_default():
        network.post_session_created()
        summary.post_session_created()
        session.post_session_created()

    with session.as_default():
        network.load()
        for i in tqdm(range(steps)):
            network.train()
            if i % summary_freq == 0 and i > 0:
                summary.summary()
                summary.flush()
            if i % save_freq == 0 and i > 0:
                network.save()

    with session.as_default():
        network.save()


def train_dist(definition_func):
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config('dxln.yml')


@configurable(config, with_name=True)
def train_task_dist(cluster, name, dataset_maker_name, network_maker_name, summary_maker_name, job_name, task_index, id_gpu=None):
    import os
    if id_gpu is not None:
        os.environ['CUDA_VISIBLE_DEVICES'] = '"{}"'.format(id_gpu)
    from dxpy.learn.distribute.cluster import get_server
    from dxpy.learn.config import config as learn_config
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    server = get_server(cluster, job_name, task_index, config=config)
    learn_config['is_chief'] = (task_index == 0)
    if job_name == 'ps':
        server.join()
    elif job_name == 'worker':
        with tf.device(tf.train.replica_device_setter(ps_device='/job:ps/cpu:0',
                                                      worker_device="/job:worker/task:{}".format(
                                                          task_index),
                                                      cluster=cluster)):
            _, network, _ = create_dataset_network_summary(dataset_maker_name,
                                                           network_maker_name,
                                                           summary_maker_name)

            train_with_monitored_session(network,
                                         target=server.target,
                                         is_chief=(task_index == 0))


def create_dataset_network_summary(dataset_maker_name, network_maker_name, summary_maker_name):
    from dxpy.learn.dataset.api import get_dataset
    from dxpy.learn.net.api import get_network, get_summary
    from dxpy.learn.utils.general import pre_work
    pre_work()
    dataset = get_dataset(dataset_maker_name)
    network = get_network(network_maker_name, dataset=dataset)
    result = network()
    summary = get_summary(summary_maker_name, dataset, network, result)
    return dataset, network, summary


def train_task_local(dataset_maker_name, network_maker_name, summary_maker_name):
    _, network, _ = create_dataset_network_summary(dataset_maker_name,
                                                   network_maker_name,
                                                   summary_maker_name)
    train_with_monitored_session(network)


def train_with_monitored_session(network, is_chief=True, target=None, steps=10000000000000):
    from dxpy.learn.utils.general import pre_work
    from dxpy.learn.session import set_default_session
    from tqdm import tqdm
    import time
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    hooks = []
    # hooks.append(tf.train.StepCounterHook())
    trainer = network['trainer']
    if 'sync_hook' in trainer.nodes:
        sync_hook = trainer['sync_hook']
        hooks.append(sync_hook)
    from dxpy.debug.utils import dbgmsg
    dbgmsg(hooks)
    with tf.train.MonitoredTrainingSession(master=target, config=config, checkpoint_dir='./save', hooks=hooks, is_chief=is_chief) as sess:
        dbgmsg('SESS CREATED')
        set_default_session(sess)
        dbgmsg('BEFORE RESET')
        network.nodes['trainer'].run('set_learning_rate')
        dbgmsg('LR RESET')
        for _ in tqdm(range(steps)):
            network.train()
